#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageRegistrationMethod.h"
#include "itkRigid2DTransform.h"
#include "itkNormalizedCorrelationImageToImageMetric.h"
#include "itkLinearInterpolateImageFunction.h"
#include "itkFRPROptimizer.h"
#include "itkCommand.h"

#include <getopt.h>
#include <iostream>

struct arguments
{
   std::string  fixImageFile;  /* -f option */
   std::string  movImageFile;  /* -m option */

   arguments () :
      fixImageFile(""),
      movImageFile("")
   {
   }

   friend std::ostream& operator<< (std::ostream& o,
                                    const arguments& args)
   {
      return o
         <<"Arguments structure:"<<std::endl
         <<"  Fixed  image file: "<<args.fixImageFile<<std::endl
         <<"  Moving image file: "<<args.movImageFile<<std::endl;
   }
};

static const char *optString = "f:m:h?";

static const struct option longOpts[] = {
   { "fix-image", required_argument, NULL, 'f' },
   { "mov-image", required_argument, NULL, 'm' },
   { "help", no_argument, NULL, 'h' },
   { NULL, no_argument, NULL, 0 }
};

/* Display program usage, and exit.
 */
void display_usage( const std::string progname )
{   
   struct arguments defargs = arguments();
   
   std::cout<<std::endl;
   std::cout<<progname<<" - Process an image"<<std::endl;
   std::cout<<"Usage: "<<progname<<" [OPTION...]"<<std::endl;
   
   std::cout<<"  -f/--fix-image=STRING     Fixed image filename - mandatory"
            <<std::endl;
   std::cout<<"  -m/--mov-image=STRING     Moving image filename - mandatory"
            <<std::endl;
   
   std::cout<<"  -h/--help                 Display this message and exit"
            <<std::endl;

   std::cout<<std::endl;
   
	exit( EXIT_FAILURE );
}



void parseOpts (int argc, char **argv, struct arguments & args)
{
   const std::string progname("ITKTestProgram");

   if (argc == 1)
   {
      display_usage(progname);
   }

   // Default values.
   args = arguments();
   
   int opt = 0; /* it's actually going to hold a char */
   int longIndex = 0;

   while ( (opt = getopt_long(argc, argv, optString,
                              longOpts, &longIndex)) != -1 )
   {
      switch( opt ) {
      case 'f':
         if (! optarg) display_usage(progname);
         args.fixImageFile = optarg;
         break;

      case 'm':
         if (! optarg) display_usage(progname);
         args.movImageFile = optarg;
         break;

      case '?':   /* fall-through is intentional */
         std::cout<<"Invalid option"<<std::endl<<std::endl;
      case 'h':	/* fall-through is intentional */
      default:
         display_usage(progname);
         break;
		}
   }

   // Check if the user provided the input images
   if ( args.fixImageFile.empty() )
   {
      std::cout<<"You need to provide a fixed image"
               <<std::endl<<std::endl;
      display_usage(progname);
   }

   if ( args.movImageFile.empty() )
   {
      std::cout<<"You need to provide a moving image"
               <<std::endl<<std::endl;
      display_usage(progname);
   }
}




namespace itk
{
/** \class OptimizerCommandIterationUpdate
//  The following section of code implements a Command observer
//  used to monitor the evolution of the registration process.
 */
class OptimizerCommandIterationUpdate : public itk::Command 
{
public:
   typedef  OptimizerCommandIterationUpdate   Self;
   typedef  itk::Command             Superclass;
   typedef  itk::SmartPointer<Self>  Pointer;
   itkNewMacro( Self );
protected:
   OptimizerCommandIterationUpdate() {};
public:
   typedef   const itk::FRPROptimizer  *    FRPROptimizerPointer;
   
   void Execute(Object *caller, const itk::EventObject & event)
   {
      Execute( (const itk::Object *)caller, event);
   }
   
   void Execute(const itk::Object * object, const itk::EventObject & event)
   {
      FRPROptimizerPointer frproptimizer = 
         dynamic_cast< FRPROptimizerPointer >( object );

      if( ! itk::IterationEvent().CheckEvent( &event ) )
      {
         return;
      }
      
      if ( frproptimizer )
      {
         //if ( frproptimizer->GetCurrentLineIteration()>0 )
         //   return;
         
         std::cout << frproptimizer->GetCurrentIteration() << "   ";
         std::cout << frproptimizer->GetValue() << "   ";
         if (frproptimizer->GetCurrentPosition().Size()<7)
            std::cout << frproptimizer->GetCurrentPosition() << "   ";
         std::cout << std::endl;
      }
   }
};

} // namespace itk




template <class InputPixelType, unsigned int Dimension>
void ProcessingFunction( arguments args )
{
   // Some typedefs
   typedef itk::Image< InputPixelType, Dimension > ImageType;

   typedef itk::Point<double,Dimension> PointType;
   typedef itk::Vector<double,Dimension> SpacingType;
   typedef itk::ImageRegion<Dimension> RegionType;
   typedef itk::Matrix<double,Dimension,Dimension> DirectionType;
 
   // Set up the file readers
   typedef typename itk::ImageFileReader< ImageType > ImageReaderType;

   typename ImageReaderType::Pointer imageReader
      = ImageReaderType::New();

   // Get the input image
   imageReader->SetFileName( args.fixImageFile.c_str() );

   // Update the reader
   try
   {
      imageReader->Update();
   }
   catch( itk::ExceptionObject& err )
   {
      std::cout << "Could not read the fixed image." << std::endl;
      std::cout << err << std::endl;
      exit( EXIT_FAILURE );
   }

   typename ImageType::Pointer fixImage = imageReader->GetOutput();
   fixImage->DisconnectPipeline();


   // Get the reference image   
   imageReader->SetFileName( args.movImageFile.c_str() );

   // Update the reader
   try
   {
      imageReader->Update();
   }
   catch( itk::ExceptionObject& err )
   {
      std::cout << "Could not read the moving image." << std::endl;
      std::cout << err << std::endl;
      exit( EXIT_FAILURE );
   }

   typename ImageType::Pointer movImage = imageReader->GetOutput();
   movImage->DisconnectPipeline();

   
   //Do the processing here
   typedef itk::ImageRegistrationMethod
      <ImageType, ImageType>                        RegistrationType;
   typedef itk::Rigid2DTransform<double>            TransformType;
   typedef itk::NormalizedCorrelationImageToImageMetric
      <ImageType, ImageType>                        MetricType;
   typedef itk::LinearInterpolateImageFunction
      <ImageType,double>                            InterpolatorType;
   typedef itk::FRPROptimizer                       OptimizerType;
   

   //-----------------------------------------------------------
   // Connect components of the registration process
   //-----------------------------------------------------------
   typename RegistrationType::Pointer registration=RegistrationType::New();
   typename TransformType::Pointer    transform   =TransformType::New();
   typename MetricType::Pointer       metric      =MetricType::New();
   typename InterpolatorType::Pointer interpolator=InterpolatorType::New();
   OptimizerType::Pointer             optimizer   =OptimizerType::New();

   // FRPR
   optimizer->SetMaximumIteration( 100 );
   optimizer->SetStepLength( 1.0 );
   optimizer->SetStepTolerance( 1e-3 );
   optimizer->SetValueTolerance( 1e-6 );
   OptimizerType::ScalesType scales(3);
   scales[0] = 1e2;
   scales[1] = 1e-2;
   scales[2] = 1e-2;
   optimizer->SetScales(scales);

   itk::OptimizerCommandIterationUpdate::Pointer observer =
      itk::OptimizerCommandIterationUpdate::New();
   optimizer->AddObserver( itk::IterationEvent(), observer );

   //metric
   metric->SubtractMeanOn();
   


   registration->SetFixedImage(  fixImage );
   registration->SetMovingImage( movImage );
   registration->SetFixedImageRegion( fixImage->GetLargestPossibleRegion() );
   registration->SetTransform( transform );
   registration->SetMetric(    metric );
   registration->SetInterpolator(  interpolator  );
   registration->SetOptimizer( optimizer );
   



   typedef typename RegistrationType::ParametersType ParametersType;
   ParametersType initialParameters( transform->GetNumberOfParameters() );
   
   initialParameters[0] = 0.0;
   initialParameters[1] = 0.0;
   initialParameters[2] = 0.0;
   
   registration->SetInitialTransformParameters( initialParameters );


   // Do the registration
   try 
   { 
      registration->StartRegistration(); 
   } 
   catch( itk::ExceptionObject & err ) 
   { 
      std::cerr << "ExceptionObject caught !" << std::endl; 
      std::cerr << err << std::endl; 
      return;
   } 
   
   ParametersType finalParameters =
      registration->GetLastTransformParameters();

   unsigned int numberOfIterations = optimizer->GetCurrentIteration();
  
   double bestValue = optimizer->GetValue();


   std::cout << std::endl;
   std::cout << "Result = " << std::endl;
   std::cout << " Transform    = " << finalParameters  << std::endl;
   std::cout << " Iterations   = " << numberOfIterations << std::endl;
   std::cout << " Metric value = " << bestValue          << std::endl;
}


int main( int argc, char *argv[] )
{
   struct arguments args;
   parseOpts (argc, argv, args);

   std::cout<<"Processing with the following arguments:"<<std::endl;
   std::cout<<args<<std::endl<<std::endl;

   // Get the image dimension and pixel type
   itk::ImageIOBase::Pointer imageIO =
      itk::ImageIOFactory::CreateImageIO(
         args.fixImageFile.c_str(), itk::ImageIOFactory::ReadMode);

   if (not imageIO)
   {
      std::cout << "Could not create a valid ImageIO for the input file."
                << std::endl;
      exit( EXIT_FAILURE );
   }
   
   imageIO->SetFileName(args.fixImageFile.c_str());

   try
   {
      imageIO->ReadImageInformation();
   }
   catch( itk::ExceptionObject& err )
   {
      std::cout << "Could not read the input image information."
                << std::endl;
      std::cout << err << std::endl;
      exit( EXIT_FAILURE );
   }
   

   // Check if the image is a scalar field
   if ( imageIO->GetPixelType() != itk::ImageIOBase::SCALAR )
   {
      std::cerr << "Unsuported input image pixel type" << std::endl;
      exit( EXIT_FAILURE );
   }

   std::cout<<"Input image dimension: "
            <<imageIO->GetNumberOfDimensions()<<std::endl;
   std::cout<<"Input image component type: "
            <<imageIO->GetComponentTypeAsString(
               imageIO->GetComponentType())<<std::endl;

   switch ( imageIO->GetNumberOfDimensions() )
   {
   case 2:
   {
      switch ( imageIO->GetComponentType() )
      {
      case itk::ImageIOBase::SHORT:
      {
         ProcessingFunction<short, 2>(args);
         break;
      }
//       case itk::ImageIOBase::FLOAT:
//       {
//          ProcessingFunction<float, 2>(args);
//          break;
//       }
      default:
      {
         // Fallback to a default IO pixeltype
         std::cout <<"Unsuported input image pixel type => Fallback" << std::endl;
         ProcessingFunction<short, 2>(args);
         break;
      }
      }
      break;
   }
   default:
   {
      std::cerr << "Unsuported input image dimension" << std::endl;
      exit( EXIT_FAILURE );
   }
   }
	
	return EXIT_SUCCESS;
}
